; GOBLK.LSP

; Written by Sri Vaidyanathan
; Bechtel Corp.
; 50 Beale St., PO Box 3965,
; San Francisco, CA 94119-3965, USA
; 415-768-1234
; 11/18/91


; Place a user-specifed block on a line, arc, or polyline and break that element
; at the edges of the block.  Other elements that cross through the block
; can also be broken at block's edge, if desired. The user is allowed to align
; the block with the element interactively.

; The beauty of this program is that it finds the edges of user-specified
; blocks from the blocks' own geometry; no preparation is necessary to use
; blocks with this program.

; With some strange-shaped blocks this program may not break exactly
; at the block's edge.  In these cases some followup trimming may be
; necessary.

; ----------------------------------------------

;error handler

(defun *error* (s)
(if (> (length vars) 0) (progn
    (if (= s "") (command "layer" "s" (nth 0 vars) ""))
    (setvar "dragmode" (nth 1 vars))
    (setvar "coords" (nth 2 vars))
    (setvar "cmdecho" (nth 3 vars))
    (setvar "regenmode" (nth 4 vars))
    (setvar "attreq" (nth 5 vars))
    (setvar "osmode" (nth 6 vars))
    (setvar "attdia" (nth 7 vars))
))
(if (and (/= s "") (/= s "Function cancelled"))
    (princ (strcat "\nError: " s))
)
(princ)
)

;bsav - blk name
;e,e1,e2 - elem list
;p,p0,p1 - points
;a,a1 - angle
;s - sel set
;i - index
;u - undo counter
;scl - scale
;x1,y1 - min blk range (global)
;x2,y2 - max blk range (global)

(defun goblk ()
(setq vars (list (getvar "clayer") (getvar "dragmode") (getvar "coords")
                 (getvar "cmdecho") (getvar "regenmode") (getvar "attreq")
                 (getvar "osmode") (getvar "attdia")
))
(setvar "dragmode" 2)
(setvar "coords" 2)
(setvar "cmdecho" 0)
(setvar "regenmode" 1)
(setvar "attreq" 1)
(setvar "osmode" 0)
(setvar "attdia" 1)
(setvar "highlight" 1)

(setq scl (getvar "dimscale") s nil e1 nil e2 nil e nil p nil p0 nil p1 nil
  u nil i 1 bsav "")
(getblk)

(while (/= bsav "X")
 (setq e (entget (entlast))
       p (cdr (assoc 10 e))             ;blk origin
       p (trans p (cdr(assoc -1 e)) 1)  ;convert from ecs to ucs
 )
 (brn e)                                ;find blk range -- function defined below

 (entdel (cdr(assoc -1 e)))             ;erase blk temporarily
 (setq e1 (ssget p))                    ;sel elem below blk
 (entdel (cdr(assoc -1 e)))             ;restore blk 
 (command "redraw")

 (if e1 (progn                           ;blk placed on a elem
        (setq e1 (entget (ssname e1 0))  ;get elem list
              p1 (cdr(assoc 0 e1))       ;elem type
        )
        (if (= p1 "LINE")
            (progn
             (setq p0 (angle (cdr(assoc 10 e1)) (cdr(assoc 11 e1)) ))
             (command "rotate" "l" "" p (angtos p0))
            )
            (if (or (= p1 "ARC")(= p1 "CIRCLE")(= p1 "POLYLINE"))
             (progn
             (prompt "\nRotation angle:")
             (command "rotate" "l" "" p pause)
             )
             (progn
             (prompt "\ncannot break that entity!")
             (setq e1 nil p1 nil)
            ))
        ) ;end if p1
        ) ;end if e1 progn

        (progn                                  ;no elem found
        (setq e1 nil p1 nil) 
        (prompt "\nRotation angle:")
        (command "rotate" "l" "" p pause)
        )
 ) ;end if e1

 (setq e (entget (cdr(assoc -1 e)))             ;get elist after rot
       a (cdr(assoc 50 e)) a1 (+ a (/ pi 2))    ;rot ang
       x2 (polar p a x2)                       ;blk inters pt (right)
       x1 (polar p (+ a pi) x1)                ;blk inters pt (left)
       y2 (polar p a1 y2)                      ;blk inters pt (top)
       y1 (polar p (+ a1 pi) y1)               ;blk inters pt (bottom)
       u 1                                      ;undo counter
 )

 (if (and e1 p1)                                ;elem found
  (if (= (cdr(assoc 0 e1)) "LINE")
         (progn
         (fndint e)                             ;find inters pts p0 & p1
         (command "break" (cdr(assoc -1 e1)) p0 p1) ;break the elem below blk
         ) ;end line
      
         (progn                                 ;for other elem types
         (setq p0 (inters x1 (polar x1 a1 1) y1 (polar y1 a 1) nil) ;left botm
               p1 (inters x2 (polar x2 a1 1) y1 (polar y1 a 1) nil) ;rt botm
               p  (inters x2 (polar x2 a1 1) y2 (polar y2 a 1) nil) ;top rt
               i  (inters x1 (polar x1 a1 1) y2 (polar y2 a 1) nil) ;top lt
         )
         (if (and p0 p1 p i)                    ;draw box
             (command "pline" p0 "w" 0 0 p1 p i "c")
             (command "pline" x1 "w" 0 0 y1 x2 y2 "c")
         )    
         (setq p (entlast) u (+ u 1)            ;save box ;incr undo counter
               kin (inters x1 x2 y1 y2 nil)     ;mid pt of box
         )
         (command "trim" p "" (list (cdr(assoc -1 e1)) kin) "") ;trim the elem below blk
         (entdel p)                             ;del box
 ))) ;end if

                                                ;find blk rng mid pt for flip/swap
  (setq p0 (/ (+ (car x1) (car x2)) 2.0)       
        p1 (/ (+ (cadr x1) (cadr x2)) 2.0)
        p  (list p0 p1)
        kin T
  )
 
  (while kin
        (initget "X F S K")
        (setq kin (getkword "\nKeyin X(exit), F(flip), S(swap), K(skip), or press RETURN to accept "))
        (cond ((= kin "F")
               (command "rotate" (cdr(assoc -1 e)) "" p 180)
               (setq u (+ u 1)  e (entget (cdr(assoc -1 e)) ))  ;get new elem list
              )
              ((= kin "S")
               (setq a (+ (cdr(assoc 50 e)) (/ pi 2)))          ;rot+90 deg 
               (command "mirror" (cdr(assoc -1 e)) "" p (polar p a 1) "y")
               (setq u (+ u 1)  e (entget (cdr(assoc -1 e)) ))  ;get new elem list
              )
              ((= kin "X") (setq bsav kin  kin nil))            ;exit
              ((= kin "K") (command "undo" u) (entdel(cdr(assoc -1 e))) (setq kin nil))  ;skip
        ) ;end cond
  ) ;end while kin

 (setq s nil e1 nil e2 nil e nil p nil p0 nil p1 nil u nil i nil)
 (if (/= bsav "X") (getblk))
) ;end while 

(*error* "")    ;reset sys vars
) ;end func


;get block name to insert

(defun getblk ()
(setq p "?")
(while (= p "?")
 (initget "? X S")
 (setq p (getpoint (strcat "\n<" (if i "select" "insert") " block> " bsav
                   (if i "" "/ Scale") "/ press RETURN to " (if i "keyin Name" "sel blk")
                   "/ ? for list/ eXit: ")))
 (cond ((= p "?") (command "block" "?" (command))) ;list blks

       ((= p "X") (setq bsav p))                   ;exit

       ((and (= p "S")(null i))                    ;scale
        (initget 6) 
        (setq p (getdist (strcat "\nScale factor <" (rtos scl 2 2) ">:")))
        (if p (setq scl p))
        (setq p "?")
       ) ;end S

       ((null p) (if i (progn                      ;keyin blk name
                        (setq p (getstring (strcat "\nBlock name <" bsav ">: ")))
                        (if (= p "") (setq p bsav))
                        (if (or (tblsearch "block" p) (findfile (strcat p ".dwg")))
                            (setq bsav p i nil p "?")
                            (progn (setq p "?")(prompt " not found!"))
                        ) ;end if tbl
                       ) ;end if i progn
                       (setq i 1 p "?")            ;select blk 
                 ) ;end if i
       ) ;end cond p

       (T        (if i (progn                      ;select blk @ given pt
                        (setq s (ssget p))   
                        (if s (setq s (entget (ssname s 0))))
                        (if (= (cdr(assoc 0 s)) "INSERT")
                            (setq bsav (cdr(assoc 2 s)) i nil p "?")
                            (progn (setq p "?")(prompt "\nnot a block!"))
                        ) ;end if ins
                       ) ;end if i progn
                       (command "insert" bsav p scl "" 0)
                 ) ;end if i
       ) ;end cond T

 ) ;end cond
) ;end while

) ;end func


;function "brn" will find blk range min & max x1,y1  x2,y2 (global vars)
;this is called from any program passing the Block entity list
;by sri 5/16/91

;e,e2 - entity list
;p,p0,p1 - points
;a,a1 - angle
;i - index

(defun brn (e / e2 p p0 p1 a a1 i) 
 (setq x1 99999 x2 -99999   ;min & max blk range   
       y1 99999 y2 -99999)  ;min & max blk range   

 (command "undo" "mark")
 (setq e2 (cdr (assoc -1 e))) ;get blk name
 (command "explode" e2)       ;expld blk

 (while (setq e2 (entnext e2))    ;exam exploded elems
    (setq e2 (entget e2)          ;elem list
          p1 (cdr(assoc 0 e2)))   ;elem type
    (cond
      ((= p1 "POINT")  
        (setq p (cdr(assoc 10 e2))
              p (trans p (cdr(assoc -1 e2)) 1)  ;convert from ecs to ucs
              x1 (min x1 (car p))
              x2 (max x2 (car p))
              y1 (min y1 (cadr p))
              y2 (max y2 (cadr p))
      ))

      ((= p1 "LINE")
        (setq p (cdr(assoc 10 e2))              ;start pt
              p (trans p (cdr(assoc -1 e2)) 1)  ;convert from ecs to ucs
              x1 (min x1 (car p))
              x2 (max x2 (car p))
              y1 (min y1 (cadr p))
              y2 (max y2 (cadr p))
              p (cdr(assoc 11 e2))              ;end pt
              p (trans p (cdr(assoc -1 e2)) 1)  ;convert from ecs to ucs
              x1 (min x1 (car p))
              x2 (max x2 (car p))
              y1 (min y1 (cadr p))
              y2 (max y2 (cadr p))
      ))

      ((= p1 "CIRCLE")
       (SETQ p0 (cdr(assoc 10 e2))               ;center
             p0 (trans p0 (cdr(assoc -1 e2)) 1)  ;convert from ecs to ucs
             p1 (cdr(assoc 40 e2))               ;radius
             a 0 a1 (/ pi 2)
       )
       (repeat 4                                 
        (SETQ p (POLAR p0 a p1)
             x1 (min x1 (car p))
             x2 (max x2 (car p))
             y1 (min y1 (cadr p))
             y2 (max y2 (cadr p))
             a (+ a a1)                         ;find pts at 90 deg intervals
      )))

      ((= p1 "ARC")
        (setq p0 (cdr(assoc 10 e2))               ;center
              p0 (trans p0 (cdr(assoc -1 e2)) 1)  ;convert from ecs to ucs
              p1 (cdr(assoc 40 e2))               ;radius
              A (cdr(assoc 50 e2))                ;start angle
              A1 (cdr(assoc 51 e2))               ;end angle
              i (abs (- a a1))                   ;included angle
        )
        (if (< a1 a) (setq i (- (* 2 pi) i) ))    ;adj incl ang for direction
        (setq a1 (+ a i) i (/ pi 180))            
        (while (<= a a1)                          ;trace the arc @1 deg interval
         (SETQ p (POLAR p0 a p1 )
              x1 (min x1 (car p))
              x2 (max x2 (car p))
              y1 (min y1 (cadr p))
              y2 (max y2 (cadr p))
               a (+ a i)
        ))
      )

      ((= p1 "INSERT")
       (command "explode" (cdr(assoc -1 e2)))   ;explode nested blk
      )

      ((= p1 "POLYLINE")
       (command "explode" (cdr(assoc -1 e2)))   ;explode elem
      )

;     (T (command "erase" (cdr(assoc -1 e2)) "")  ;ignore other elem types
;     )

    ) ;end cond
    (setq e2 (cdr(assoc -1 e2)))                  ;entity name
 ) ;end while

 (command "undo" "back")
 (setq p (cdr (assoc 10 e))             ;blk origin
       p (trans p (cdr(assoc -1 e)) 1)  ;convert from ecs to ucs
 )

 (setq p0 (polar p 0 1))                               ;end pt of imaginary line thru blk org
 (if (= y1 y2) (setq p1 (+ y1 1)) (setq p1 y2))        ;ensure inters
 (setq i (inters p p0 (list x1 y1) (list x1 p1) nil)   ;left inters pt
       x1 (distance p i)                               ;offset from blk org
       i (inters p p0 (list x2 y1) (list x2 p1) nil)   ;right inters pt
       x2 (distance p i)                               ;offset from blk org
 )

 (setq p0 (polar p (/ pi 2) 1))                        ;end pt of imaginary line thru blk org
 (if (= x1 x2) (setq p1 (+ x1 1)) (setq p1 x2))        ;ensure inters
 (setq i (inters p p0 (list x1 y1) (list p1 y1) nil)   ;bottom inters pt
       y1 (distance p i)                               ;offset from blk org
       i (inters p p0 (list x1 y2) (list p1 y2) nil)   ;top inters pt
       y2 (distance p i)                               ;offset from blk org
 )

) ;end func

; function "fndint" will find exact intersection of block & the imaginary line
; thru its origin representing the block range
; this pgm is called after brn.lsp
; p0,p1 - intersection pts (global)
; x1,y1 - imag line start pts (global)
; x2,y2 - imag line end pts (global)
; a,a1 - elem lists
; s - sel set
; r - dist
; j - interval

(defun fndint (a / a1 s r j)
(setq s (ssadd))                        ;create null sel set
(ssadd (cdr(assoc -1 a)) s)             ;add blk name

(setq r (distance x1 x2))
(cond ((> r 0.125) (setq j 0.125))      ;1/8"  intervals for obj snap
      ((> r 0.0625) (setq j 0.0625))    ;1/16"
      (T (setq j 0.03125))              ;1/32"
)

(setq p0 (intpt x1 (angle x1 x2) s)     ;find left inters pt
      p1 (intpt x2 (angle x2 x1) s))    ;find right inters pt

(if (and p0 p1) ()                      ;found p0 & p1, so quit
(progn                                  ;else

 (if (or (< (cdr(assoc 41 a)) 0) (< (cdr(assoc 42 a)) 0) (< (cdr(assoc 43 a)) 0))
     (progn                              ;negative scale factor, can't explode blk,
                                         ; so use orig inters pt
     (if (null p0) (setq p0 x1))    
     (if (null p1) (setq p1 x2))
     )

     (progn                                 ;else explode blk
     (setq a1 (entlast))
     (command "explode" (cdr(assoc -1 a)))
     (while (setq a1 (entnext a1))          ;exam exploded elems
        (setq a1 (entget a1)                ;elem list
              i (cdr(assoc 0 a1)))          ;elem type
        (if (or (= i "POINT")(= i "LINE")(= i "ARC")
                (= i "CIRCLE")(= i "POLYLINE")(= i "INSERT"))
            (ssadd (cdr(assoc -1 a1)) s)    ;add elem to sel set
        )
        (setq a1 (cdr(assoc -1 a1)))        ;elem name
     )
     (if (null p0) (setq p0 (intpt x1 (angle x1 x2) s)))    ;find left inters pt
     (if (null p1) (setq p1 (intpt x2 (angle x2 x1) s)))    ;find right inters pt
     (command "u")                           ;undo explode
     (if (null p0) (setq p0 x1))             ;use orig inters pt as last attempt
     (if (null p1) (setq p1 x2))    

     ) ;end progn
 ) ;end if neg scale

)) ;end if p0 & p1 found
) ;end func


;trace line path @j intervals and find inters pt
;p2 - start pt
;a1 - line angle
;s,s1 - sel set
;i - index
;j - interval
;k - intermediate pt

(defun intpt (p2 a1 s / s1 i k)
(setq i 0)
(while (<= i r)
 (setq k (osnap p2 "int,near"))
 (if k (setq s1 (ssget k))            ;find elem at inters pt
       (setq s1 (ssget p2))            ;else find elem at this pt
 )
 (if (and s1 (ssmemb (ssname s1 0) s)) ;check if the inters elem is part of the block
     (progn
     (if k (setq p2 k))                  ;osnap pt
     (setq i (+ i r))                    ;stop !
     )
     (progn                              ;else
     (setq p2 (polar p2 a1 j)            ;get next pt      
            i (+ i j)                    ;incr dist counter
           s1 nil)
     )
 )
) ;end while

(if s1 p2 nil)                          ;return inters pt
) ;end func





